@@ -1,32 +1,39 @@
 # -*- coding: utf-8 -*-
-__version__ = "0.1.1"
+__version__ = "1.0.0"
 
 import re
-from ConfigParser import ConfigParser
-from itertools import product
-from fnmatch import fnmatch
-
 try:
     from collections import OrderedDict
 except ImportError:
     from .ordereddict import OrderedDict
+try:
+    from ConfigParser import ConfigParser
+except ImportError:
+    from configparser import ConfigParser
+from fnmatch import fnmatch
+from itertools import product
+import warnings
+
+try:
+    from StringIO import StringIO
+except ImportError:
+    from io import StringIO
 
 
 entry_rx = re.compile(r"""
     ^
-    ((?P<alias>[^:]*):)?
+    ((?P<merge>\?))?
+    ((?P<alias>[^\?:]*):)?
     \s*(?P<value>[^!&]+?)\s*
     (?P<reducers>[!&].+)?
     $
 """, re.VERBOSE)
 reducer_rx = re.compile(r"""
-
-        \s*
-        (?P<type>[!&])
-        (?P<variable>[^!&\[\]]+)
-        \[(?P<glob>[^\[\]]+)\]
-        \s*
-
+    \s*
+    (?P<type>[!&])
+    (?P<variable>[^!&\[\]]+)
+    \[(?P<glob>[^\[\]]+)\]
+    \s*
 """, re.VERBOSE)
 
 special_chars_rx = re.compile(r'[\\/:>?|\[\]< ]+')
@@ -36,14 +43,21 @@ class ParseError(Exception):
     pass
 
 
-class DuplicateEntry(ParseError):
+class DuplicateEntry(UserWarning):
     def __str__(self):
-        return "Duplicate entry %r (from %r). Conflicts with %r - it has the same alias."
+        return "Duplicate entry %r (from %r). Conflicts with %r - it has the same alias." % self.args
+    __repr__ = __str__
+
+
+class DuplicateEnvironment(Exception):
+    def __str__(self):
+        return "Duplicate environment %r. It has conflicting sets of data: %r != %r." % self.args
     __repr__ = __str__
 
 
 class Reducer(object):
-    def __init__(self, (kind, variable, pattern)):
+    def __init__(self, entry):
+        kind, variable, pattern = entry
         assert kind in "&!"
         self.kind = kind
         self.is_exclude = kind == '!'
@@ -65,6 +79,7 @@ class Entry(object):
         if not value or value == '-':
             self.alias = ''
             self.value = ''
+            self.merge = False
             self.reducers = []
         else:
             m = entry_rx.match(value)
@@ -73,6 +88,7 @@ class Entry(object):
             m = m.groupdict()
             self.alias = m['alias']
             self.value = m['value']
+            self.merge = m['merge']
             self.reducers = [Reducer(i) for i in reducer_rx.findall(m['reducers'] or '')]
 
         if self.alias is None:
@@ -90,9 +106,9 @@ class Entry(object):
     __repr__ = __str__
 
 
-def parse_config(filename):
+def parse_config(fp):
     parser = ConfigParser()
-    parser.read(filename)
+    parser.readfp(fp)
     config = OrderedDict()
     for name, value in parser.items('matrix'):
         entries = config[name] = []
@@ -100,9 +116,10 @@ def parse_config(filename):
             entry = Entry(line)
             duplicates = [i for i in entries if i == entry]
             if duplicates:
-                raise DuplicateEntry(entry, line, duplicates)
-            else:
-                entries.append(entry)
+                warnings.warn(DuplicateEntry(entry, line, duplicates), DuplicateEntry, 1)
+            entries.append(entry)
+        if not entries:
+            entries.append(Entry('-'))
     return config
 
 
@@ -121,9 +138,13 @@ def from_config(config):
                 if match if reducer.is_exclude else not match:
                     include = False
         if include:
-            matrix['-'.join(entry.alias for entry in entries if entry.alias)] = dict(
+            key = '-'.join(entry.alias for entry in entries if entry.alias)
+            data = dict(
                 zip(variables, (entry.value for entry in entries))
             )
+            if key in matrix and data != matrix[key]:
+                raise DuplicateEnvironment(key, data, matrix[key])
+            matrix[key] = data
     return matrix
 
 
@@ -131,5 +152,13 @@ def from_file(filename):
     """
     Generate a matrix from a .ini file. Configuration is expected to be in a ``[matrix]`` section.
     """
-    config = parse_config(filename)
+    config = parse_config(open(filename))
+    return from_config(config)
+
+
+def from_string(string):
+    """
+    Generate a matrix from a .ini file. Configuration is expected to be in a ``[matrix]`` section.
+    """
+    config = parse_config(StringIO(string))
     return from_config(config)
